# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import DeleteTriggerResult


class DeleteTrigger(GQLBase):
    delete_trigger: Optional[DeleteTriggerResult] = Field(alias="deleteTrigger")


DeleteTrigger.model_rebuild()
