# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import UpdateFilterTriggerResult


class UpdateFilterTrigger(GQLBase):
    update_filter_trigger: Optional[UpdateFilterTriggerResult] = Field(
        alias="updateFilterTrigger"
    )


UpdateFilterTrigger.model_rebuild()
